import json
from typing import List, Optional
from .query_result import QueryResult, QuadStore
from .anzo_request import AnzoRequestBuilder
from .uris import ALL_NAMED_GRAPHS_URI
# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/


class AnzoClient:
    """A class for interacting with an Anzo Server.

    This wraps the anzo HTTP endpoints in order to ask questions of Anzo
    and invoke actions on the Anzo server.

    The general patterns throughout this class are:

    - ValueError is raised if the arguments are invalid
    - Arguments are checked frequently and strictly
    - An error is raised if the operation was unsuccessful (e.g., a query
      with a syntax error)
        - A TimeoutError is raised if the operation times out
        - A RuntimError is thrown otherwise
    - Query results are returned as a QueryResult object, from which the user
      can retrieve the results in whichever form is most useful

    Attributes:
        server: The server that Anzo is running on.
        port: The port on which Anzo is listening for HTTPS traffic.
            Often 443 or 8443.
        auth_token: The authentication token that is being used
        username: Username of the user with which to interact with Anzo.
        password: The password of the user
        timeout_seconds: Timeout for operations in seconds

    Usage:
        anzo_client = AnzoClient(
            "anzo.cambridgesemantics.com", "8080", "username", "password"
        )

        query_result = anzo_client.query_graphmart(
            "urn://graphmart_uri",
            query_string="SELECT * WHERE { ?s ?p ?o } LIMIT 10",
        )

        ss_result = anzo_client.execute_semantic_service("urn://service", [])
    """

    SYSTEM_TABLE_LDS_URI = "http://openanzo.org/catEntry(%5Bhttp%3A%2F%2Fcambridgesemantics.com%2Fontologies%2F2009%2F05%2FLinkedData%23AnzoXray%5D%40%5Bhttp%3A%2F%2Fcambridgesemantics.com%2Fdatasource%2FSystemTables%5D)"  # noqa

    def __init__(self,
                 server: str,
                 port: str,
                 username: str = "",
                 password: str = "",
                 auth_token: str = "",
                 timeout_seconds: float = 7200.,
                 ssl_workaround = True) -> None:
        """Constructs an AnzoClient for interacting with an Anzo server.

        Args:
            server: The server that Anzo is running on.
            port: The port on which Anzo is listening for HTTPS traffic.
                Often 443 or 8443.
            username (optional): Username of the user with which
                to interactwith Anzo.
            password (optional): The password of the user
            auth_token (optional): An authorization token.
            timeout_seconds (optional): Timeout for operations in seconds
            ssl_workaround (optional): Add lower security SSL DH cipher enabled by default for convenience.

        Returns:
            An AnzoClient object

        Note: An AnzoClient must be passed either a username
                and password or an auth_token, but not both.
        """

        self.server = server
        self.port = port
        if ssl_workaround:
            self.__ssl_config()
        
        if username and not password:
            error = "If a username is specified a password also needs to be"
            raise ValueError(error)

        if password and not username:
            error = "If password is specified a username also needs to be"
            raise ValueError(error)

        if username and password and auth_token:
            error = "Username and password specified as well as an auth_token"
            raise ValueError(error)

        if not (username or password or auth_token):
            error = "Some form of authentication must be provided."
            raise ValueError(error)

        self.username = username
        self.password = password
        self.auth_token = auth_token
        self.timeout_seconds = timeout_seconds

    def query_graphmart(self,
                        graphmart: str,
                        query_string: str = "",
                        query_file: str = "",
                        data_layers: List[str] = None,
                        skip_cache: bool = False) -> QueryResult:
        """Executes a SPARQL query against a Graphmart.

        Args:
            graphmart: URI of the graphmart to query
            query_string: (optional) SPARQL query string to execute against the
                graphmart. Either query_string or query_file must be provided,
                but not both.
            query_file: (optional) File with SPARQL query to execute against
                the graphmart. Either query_string or query_file must be
                provided, but not both.
            data_layers (optional): List of data layers to query. The uris
                of the layers should be specified in individual strings
                like ["urn://layer1", "urn://layer1"]. The default is to
                query all data layers graphs in the graphmart.
            skip_cache (optional): A boolean, when set to True skips Anzo's
                query cache to compute as opposed to look up the query's results.

        Returns:
            The result of the query in a QueryResult object.
        """
        # Check if either a query_string or a query_file, but not both.
        if not query_string and not query_file:
            raise ValueError(
                "Method query_graphmart requires either a query string "
                "or query file"
            )

        if query_string and query_file:
            raise ValueError(
                "Method query_graphmart received both a query string "
                "and a query file. Only one should be provided"
            )

        if not graphmart:
            raise ValueError("A graphmart is required")

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_graphmart(graphmart, data_layers)

        if query_string:
            arb.with_query_string(query_string)
        elif query_file:
            arb.with_query_file(query_file)

        arb.with_cache_skipped(skip_cache)

        anzo_request = arb.build()
        response_text = anzo_request.execute_request()
        return QueryResult(response_text)

    def query_journal(self, query_string: str = "",
                      query_file: str = "",
                      named_graphs: List[str] = None) -> QueryResult:
        """Executes a SPARQL query against the Anzo journal.

        Given some query, execute a query against the Anzo journal.
        Either a query_string or query_file must be provided, but not both.

        Args:
            query_string: (optional) SPARQL query string to execute against the
                journal. Either query_string or query_file must be provided,
                but not both.
            query_file: (optional) File with SPARQL query to execute against
                the journal. Either query_string or query_file must be
                provided, but not both.
            named_graphs  (optional): List of named graphs to query. The graphs
                should be specified in individual strings like
                ["urn://graph1", "urn://graph2"]. The default is to
                query all named graphs in the journal.

        Returns:
            The result of the query in a QueryResult object.
        """

        if not query_string and not query_file:
            raise ValueError(
                "Method query_journal requires either a query string "
                "or query file"
            )

        if query_string and query_file:
            raise ValueError(
                "Method query_journal received both a query string "
                "and a query file. Only one should be provided"
            )

        if not named_graphs:
            named_graphs = [ALL_NAMED_GRAPHS_URI]

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_query_journal(named_graphs)

        if query_string:
            arb.with_query_string(query_string)
        elif query_file:
            arb.with_query_file(query_file)

        anzo_request = arb.build()
        response_text = anzo_request.execute_request()
        return QueryResult(response_text)

    def update_journal(self, query_string: str = "",
                       query_file: str = "") -> QueryResult:
        """Executes a SPARQL update query against the Anzo journal.

        The query must be conform the UPDATE query spec, as outlined in
        https://www.w3.org/TR/sparql11-update/.

        In summary, it's expected the query have one of the following clauses:
        INSERT, DELETE, INSERT DATA, DELETE DATA. Other clauses, which
        are outlined in the spec linked above, are supported but not
        as well tested.

        Args:
            query_string: (optional) SPARQL query string to execute against the
                journal. Either query_string or query_file must be provided,
                but not both.
            query_file: (optional) File with SPARQL update query to execute
                against the journal. Either query_string or query_file must be
                provided, but not both.

        Returns:
            The result of the query in a QueryResult object.
        """

        if not query_string and not query_file:
            raise ValueError(
                "Method query_journal requires either a query string "
                "or query file"
            )

        if query_string and query_file:
            raise ValueError(
                "Method query_journal received both a query string "
                "and a query file. Only one should be provided"
            )

        named_graphs = [ALL_NAMED_GRAPHS_URI]

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_update_journal(named_graphs)

        if query_string:
            arb.with_update_string(query_string)
        elif query_file:
            arb.with_update_file(query_file)

        anzo_request = arb.build()
        response_text = anzo_request.execute_request()
        return QueryResult(response_text)

    def query_lds(self, lds_cat_entry: str, query_string: str = "",
                  query_file: str = "",
                  named_graphs: Optional[List[str]] = None
                  ) -> QueryResult:
        """Executes a SPARQL query against an Anzo linked dataset catalog entry

        Args:
            lds_cat_entry: URI of the linked dataset catalog entry to query.
                This URI will encode both the linked dataset and the associated
                volume datasource.
            query_string: (optional) SPARQL query string to execute against the
                lds_cat_entry. Either query_string or query_file must be
                provided, but not both.
            query_file: (optional) File with SPARQL query to execute against
                the lds_cat_entry. Either query_string or query_file must be
                provided, but not both.
            named_graphs  (optional): List of addiontal named graphs to query
                that are not already in the specified LDS. The graphs
                should be specified in individual strings like
                ["urn://graph1", "urn://graph2"].

        Returns:
            The result of the query in a QueryResult object.
        """

        if not query_string and not query_file:
            raise ValueError(
                "Method query_lds requires either a query string "
                "or query file"
            )

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_lds_cat_entry(lds_cat_entry, named_graphs)

        if query_string:
            arb.with_query_string(query_string)
        elif query_file:
            arb.with_query_file(query_file)

        anzo_request = arb.build()
        response_text = anzo_request.execute_request()
        return QueryResult(response_text)

    def query_system_tables(self,
                            query_string: str = "",
                            query_file: str = "",
                            ) -> QueryResult:
        """Executes a SPARQL query against the anzo system tables

        Args:
            query_string: (optional) SPARQL query string to execute against the
                system tables. Either query_string or query_file must be
                provided, but not both.
            query_file: (optional) File with SPARQL query to execute against
                the system tables. Either query_string or query_file must be
                provided, but not both.

        Returns:
            The result of the query in a QueryResult object.
        """

        return self.query_lds(
            self.SYSTEM_TABLE_LDS_URI, query_string, query_file,
            named_graphs=[ALL_NAMED_GRAPHS_URI]
        )

    def execute_semantic_service(self,
                                 service_uri: str,
                                 request: QuadStore) -> QuadStore:
        """Executes a Semantic Service on Anzo

        Args:
            service_uri: URI of the semantic service
            request: The request payload as a quadstore

        Returns:
            A QuadStore object with the contents of the response from the
            semenatic service call
        """
        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_semantic_service(service_uri, request.as_anzo_json_list())
        anzo_request = arb.build()
        response_json = json.loads(anzo_request.execute_request())
        return QuadStore.from_anzo_json_list(response_json)

    def get(self, named_graph_uri: str,
            include_metadata_graph: bool = False) -> QuadStore:
        """Gets the contents of a named graph and optionally its
        metadata graph

        If the graph doesn't exist, then a ValueError is raised.

        Args:
            named_graph_uri: URI of the named graph to get
            include_metadata_graph (optional): If true, then the metadata graph
                is included in the returned QuadStore, and otherwise
                it's not included. The value is false by default.

        Returns:
            A QuadStore with the contents of the named graph and optionally
            its metadata graph.
        """

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        arb.with_get(named_graph_uri)
        anzo_request = arb.build()
        response_json = json.loads(anzo_request.execute_request())
        quad_store = QuadStore.from_anzo_json_list(response_json)

        if not include_metadata_graph:
            quad_store = quad_store.filter(graph=named_graph_uri)

        return quad_store

    def add(self,
            quad_store: Optional[QuadStore] = None,
            trig_filename: Optional[str] = None) -> None:
        """Adds the quads in the quad_store to the system journal

        If the named graph doesn't exist, then it is created and registered.

        Args:
            quad_store (optional): Set of quads to add to the
               Anzo system journal
            trig_filename (optional): The file path to a trig file to add to
                the Anzo system journal
        Raises:
            ValueError: Exactly one of quad_store and trig_filename must be specified
                  otherwise a ValueError is raised.
        """
        if not (quad_store or trig_filename):
            err = "Either quad_store or trig_filename must be specified"
            raise ValueError(err)

        if quad_store and trig_filename:
            err = "Cannot specify both quad_store and trig_filename"
            raise ValueError(err)

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        if trig_filename:
            quad_store = QuadStore.from_trig_file(trig_filename)
        arb.with_add(quad_store.as_anzo_json_list())
        anzo_request = arb.build()
        anzo_request.execute_request()

    def remove(self,
               quad_store: Optional[QuadStore] = None,
               trig_filename: Optional[str] = None) -> None:
        """Removes the quads in the quad_store to the system journal

        If some or all of the quads don't exist, an error is not raised.

        Args:
            quad_store: Set of quads to remove from the Anzo system journal
            trig_filename (optional): The file path to a trig file to remove from
                Anzo.
        Raises:
            ValueError: Exactly one of quad_store and trig_filename must be specified
        """
        if not (quad_store or trig_filename):
            err = "Either quad_store or trig_filename must be specified"
            raise ValueError(err)

        if quad_store and trig_filename:
            err = "Cannot specify both quad_store and trig_filename"
            raise ValueError(err)

        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_timeout_seconds(self.timeout_seconds)
        if not quad_store:
            quad_store = QuadStore.from_trig_file(trig_filename)
        arb.with_remove(quad_store.as_anzo_json_list())
        anzo_request = arb.build()
        anzo_request.execute_request()
        
    ## Work around for SSL related DH Key errors
    # ex: [SSL: SSL_NEGATIVE_LENGTH] dh key too small (_ssl.c:600)
    def __ssl_config(self):
        import requests
        try:
            requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += 'HIGH:!DH:!aNULL'
            requests.packages.urllib3.contrib.pyopenssl.DEFAULT_SSL_CIPHER_LIST += 'HIGH:!DH:!aNULL'
        except AttributeError:
            # no pyopenssl support used / needed / available
            pass



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
import os
import json
import copy
import parameterized


from pyanzo import (
    AnzoClient,
    AnzoRequestBuilder,
    AnzoRequest,
    Quad,
    QueryResult,
    QuadStore,
    TableResult,
    semantic_type_to_python_object,
)

from .test_common import (
    GRAPHMART,
    SERVER,
    PORT,
    EMPTY_AUTH_TOKEN,
    USERNAME,
    PASSWORD
)


class TestAnzoClientGetGraph(unittest.TestCase):
    graphmart = GRAPHMART

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(SERVER, PORT, USERNAME, PASSWORD)

    # TODO: add a test that checks the output more carefully.
    # These graphs were just so big that I didn"t want to paste
    # the expected values.
    @unittest.skip("Not implemented yet")
    def test_simple_get_graphmart(self) -> None:
        res = self.anzo_client.get_graph(self.graphmart)
        self.assertEqual(res[0]["namedGraphUri"], self.graphmart)


class TestQueryResult(unittest.TestCase):
    """
    TODO: the names of the methods in this need to be updated. They're based
    on querying anzo/graphmarts but anzo/graphmart is never actually queried.
    """

    graphmart = GRAPHMART

    # This is the basis of the query result
    # simple_select_query = """
    # PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>  # noqa
    #
    # SELECT ?city WHERE {
    #     ?s a data:Person ;
    #         data:Person_City ?city .
    # } ORDER BY ?city LIMIT 5"""

    # We construct the result manually to avoid confounding errors
    # TODO: rename based on conventions
    QueryResult_simple_select_query = QueryResult("""{
        "head": {"vars": ["city"]},
        "results": {
            "bindings": [
                {"city": {"type": "literal", "value": "Boston"}},
                {"city": {"type": "literal", "value": "Cambridge"}},
                {"city": {"type": "literal", "value": "Medford"}},
                {"city": {"type": "literal", "value": "Somerville"}},
                {"city": {"type": "literal", "value": "Watertown"}}
            ]
        }
    }""")

    # Results
    simple_select_graphmart_query_results_list = [
        ["Boston"], ["Cambridge"], ["Medford"], ["Somerville"], ["Watertown"]
    ]

    simple_select_graphmart_query_results_dict = [
        {"city": "Boston"},
        {"city": "Cambridge"},
        {"city": "Medford"},
        {"city": "Somerville"},
        {"city": "Watertown"}
    ]

    # This is the basis of the query
    # query_with_optional = """
    #     PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>  # noqa
    #
    #     SELECT ?city ?height WHERE {
    #       ?s a data:Person ;
    #            data:Person_Name ?name ;
    #            data:Person_City ?city ;
    #            data:Person_Age ?age .
    #
    #       # person height is not a property and
    #       # will be unbound by design (for testing)
    #       OPTIONAL { ?s  data:Person_Height  ?height }
    #
    #     } ORDER BY ?city LIMIT 5
    #  """

    # We construct the result manually to avoid confounding errors
    QueryResult_query_with_optional = QueryResult("""{
        "head": {"vars": ["city", "height"]},
        "results": {
            "bindings": [
                {"city": {"type": "literal", "value": "Boston"}},
                {"city": {"type": "literal", "value": "Cambridge"}},
                {"city": {"type": "literal", "value": "Medford"}},
                {"city": {"type": "literal", "value": "Somerville"}},
                {"city": {"type": "literal", "value": "Watertown"}}
            ]
        }
    }""")

    graphmart_query_with_optional_results_list = [
        ["Boston", ""], ["Cambridge", ""], ["Medford", ""],
        ["Somerville", ""], ["Watertown", ""]
    ]

    graphmart_query_with_optional_results_dict = [
        {"city": "Boston", "height": ""},
        {"city": "Cambridge", "height": ""},
        {"city": "Medford", "height": ""},
        {"city": "Somerville", "height": ""},
        {"city": "Watertown", "height": ""},
    ]

    # Templated verions of the queries to allow for different
    # SELECT vars and LIMITs
    simple_select_query_template = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT {} WHERE {{
            ?s a data:Person ;
                data:Person_City ?city .
        }} ORDER BY ?city LIMIT {}"""  # noqa

    query_with_ints = """
        PREFIX data: <http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>

        SELECT ?name ?age WHERE {{
            ?s a data:Person ;
                data:Person_Name ?name ;
                data:Person_Age ?age2 .
                BIND(DOUBLE(?age2) as ?age)
        }} ORDER BY ?age
    """  # noqa

    record_dicts_with_ints = [
        {'name': 'Alice', 'age': 30.0},
        {'name': 'Bob', 'age': 33.0},
        {'name': 'Carlos', 'age': 50.0},
        {'name': 'Diego', 'age': 60.0},
        {'name': 'Eve', 'age': 70.0}
    ]

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(SERVER, PORT, USERNAME, PASSWORD)

    # The test docstrings are convenient when looking at the code, but
    # they make the verbose testing view look cluttered. This function
    # removes comments/descriptions from the verbose test view.
    def shortDescription(self) -> None:
        return None

    @parameterized.parameterized.expand([
        ("SELECT_CITY", QueryResult_simple_select_query),
        ("SELECT_CITY_HEIGHT", QueryResult_query_with_optional),
    ])
    def test_graphmart_query_sublist_equal_length(self,
                                                  test_name: str,
                                                  query_result: QueryResult
                                                  ) -> None:
        """Test that each sublist/solution is the same length"""

        result = query_result.as_table_results()

        self.assertTrue(
            all(
                len(solution) == len(result.as_list()[-1])
                for solution in result.as_list()
            )
        )

    @parameterized.parameterized.expand([
        ("SELECT_CITY", QueryResult_simple_select_query),
        ("SELECT_CITY_HEIGHT", QueryResult_query_with_optional),
    ])
    def test_graphmart_query_dict_values_equal_length(self,
                                                      test_name: str,
                                                      query_result: QueryResult
                                                      ) -> None:
        """Test that each sublist/solution is the same length"""

        result = query_result.as_table_results()

        self.assertTrue(
            all(len(solution.values()) == len(result.as_record_dictionaries()[-1].values())  # noqa
                for solution in result.as_record_dictionaries())
        )

    # This will create two versions of this test
    #   1) test_graphmart_query_result_length_0_LIST
    #   2) test_graphmart_query_result_length_1_DICT
    # The name comes from the first value of each tuple
    # The values of the tuple are passed to the test as parameters
    # ("LIST", QueryResult.as_list) becomes
    #   test_graphmart_query_result_length(self, "LIST", QueryResult.as_list)
    @parameterized.parameterized.expand([
        ("LIST",
         TableResult.as_list,
         simple_select_graphmart_query_results_list),
        ("DICT",
         TableResult.as_record_dictionaries,
         simple_select_graphmart_query_results_dict),

        # Need to convert iter to list in order to compare results
        ("LIST_ITER", lambda x: list(TableResult.as_list_iter(x)),
            simple_select_graphmart_query_results_list)
    ])
    def test_simple_graphmart_query(self,
                                    test_name,
                                    query_result_method,
                                    expected) -> None:
        """Test a simple query returns the correct results for each
        query result method
        """
        result = self.QueryResult_simple_select_query.as_table_results()
        self.assertEqual(query_result_method(result), expected)

    @parameterized.parameterized.expand([
        ("LIST",
         TableResult.as_list,
         graphmart_query_with_optional_results_list),
        ("DICT",
         TableResult.as_record_dictionaries,
         graphmart_query_with_optional_results_dict),

        # Need to convert iter to list in order to compare results
        ("LIST_ITER", lambda x: list(TableResult.as_list_iter(x)),
            graphmart_query_with_optional_results_list)
    ])
    def test_graphmart_optional_query(self,
                                      test_name,
                                      query_result_method,
                                      expected) -> None:
        """Test a query that contains optional bindings, unbound values should
           be marked as empty string"""
        result = self.QueryResult_query_with_optional.as_table_results()

        self.assertEqual(query_result_method(result), expected)

    def test_simple_graphmart_query_internal_iter(self) -> None:
        """Test __iter__ of query result, the query result itself
           should be iterable"""
        result = self.QueryResult_simple_select_query.as_table_results()
        self.assertTrue(
            all(self.simple_select_graphmart_query_results_list[i] == solution
                for i, solution in enumerate(result))
        )

        self.assertEqual(
            list(result),
            self.simple_select_graphmart_query_results_list,
        )

    def test_simple_construct_query(self) -> None:
        """Test to ensure that simple construct queries do not error."""
        anzo_client = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT)
        query = """
        PREFIX pyanzo:<http://cambridgesemantics.com/ont/autogen/a69a/PyAnzo_Dictionary/PyAnzo_Datasource#>
        CONSTRUCT {
            ?s a pyanzo:Person  .
        }
        WHERE {
        ?s a pyanzo:Person .
        }""" # noqa
        graphmart = GRAPHMART
        expectation_uris = set(
            ['http://csi.com/Person/0c391c0f-340d-4f08-9fa6-13a1793c35cb',
             'http://csi.com/Person/a153b0aa-5583-40ff-8710-f94c8a25df3d',
             'http://csi.com/Person/c2ee95a3-d5b1-40d8-8243-9c7d541853f7',
             'http://csi.com/Person/ffbe9982-7491-42da-9398-e4c22fab5711',
             'http://csi.com/Person/879d629b-1015-45b9-8457-7271c0f2a9f0'
             ])
        results = anzo_client.query_graphmart(graphmart, query_string=query)
        quad_store = results.as_quad_store()
        person_uris = set()
        for quad in quad_store:
            person_uris.add(quad.sub)
        self.assertEqual(expectation_uris, person_uris)

    def test_record_dicts(self) -> None:
        res = self.anzo_client.query_graphmart(
            self.graphmart,
            self.query_with_ints
        )

        rd = res.as_table_results().as_record_dictionaries()
        self.assertEqual(rd, self.record_dicts_with_ints)

class TestAnzoRequestBuilder(unittest.TestCase):
    server = SERVER
    auth_token = EMPTY_AUTH_TOKEN
    username = USERNAME
    password = PASSWORD
    port = PORT
    graphmart = GRAPHMART

    query_string = "SELECT * WHERE { ?s ?p ?o } LIMIT 10"
    query_file = simple_select_query_file = os.path.join(
        "tests", "test_assets", "simple_graphmart_query.rq"
    )

    def test_missing_url_postfix(self) -> None:
        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_auth(self.username, self.password, self.auth_token)
        self.assertRaises(ValueError, arb.build)

    def test_missing_url(self) -> None:
        arb = AnzoRequestBuilder()
        arb.with_auth(self.username, self.password, self.auth_token)
        arb.with_query_string(self.query_string)
        self.assertRaises(ValueError, arb.build)

    def test_missing_auth(self) -> None:
        arb = AnzoRequestBuilder()
        arb.with_url(self.server, self.port)
        arb.with_query_string(self.query_string)
        self.assertRaises(ValueError, arb.build)


class TestAnzoRequest(unittest.TestCase):
    server = SERVER
    port = PORT
    auth_token = EMPTY_AUTH_TOKEN
    username = USERNAME
    password = PASSWORD
    graphmart = GRAPHMART
    timeout_seconds = 120  # two minutes

    query_string = "SELECT * WHERE { ?s ?p ?o } LIMIT 10"
    query_file = simple_select_query_file = os.path.join(
        "tests", "test_assets", "simple_graphmart_query.rq"
    )

    url = f"https://{server}:{port}/sparql"
    payload = {
        "query": query_string,
        "format": "json"
    }
    headers = {}

    query_json_results = {
        "head": {
            "vars": ["p", "s", "o"]
        },
        "results": {
            "bindings": []
        }
    }

    def test_good_request(self) -> None:
        req = AnzoRequest(
            self.url,
            self.auth_token,
            self.username,
            self.password,
            self.headers,
            self.payload,
            self.timeout_seconds)
        res = req.execute_request()
        self.assertEqual(json.loads(res), self.query_json_results)

    def test_bad_host(self) -> None:
        bad_host = self.server + self.server
        bad_url = f"http://{bad_host}:{self.port}/sparql"
        req = AnzoRequest(bad_url,
                          self.username,
                          self.password,
                          self.auth_token,
                          self.headers,
                          self.payload,
                          self.timeout_seconds)
        self.assertRaises(RuntimeError, req.execute_request)

    def test_bad_port(self) -> None:
        bad_port = "8088"
        self.assertNotEqual(self.port, bad_port)

        bad_url = f"http://{self.url}:{bad_port}/sparql"
        req = AnzoRequest(bad_url,
                          self.username,
                          self.password,
                          self.auth_token,
                          self.headers,
                          self.payload,
                          self.timeout_seconds)
        self.assertRaises(RuntimeError, req.execute_request)

    def test_bad_username(self) -> None:
        bad_username = self.username + self.username

        req = AnzoRequest(self.url,
                          bad_username,
                          self.password,
                          self.auth_token,
                          self.headers,
                          self.payload,
                          self.timeout_seconds)
        self.assertRaises(RuntimeError, req.execute_request)

    def test_bad_password(self) -> None:
        bad_password = self.password + self.password

        req = AnzoRequest(self.url,
                          self.username,
                          bad_password,
                          self.auth_token,
                          self.headers,
                          self.payload,
                          self.timeout_seconds)
        self.assertRaises(RuntimeError, req.execute_request)


class TestSemanticTypeToPythonObject(unittest.TestCase):
    def test_unknown_type(self) -> None:
        val = "test"
        res = semantic_type_to_python_object(
            val, "http://www.w3.org/2001/XMLSchema#unknowntype"
        )
        self.assertEqual(val, res)

    def test_string(self) -> None:
        val = "test"
        res = semantic_type_to_python_object(val, "")
        self.assertEqual(val, res)

    def test_int(self) -> None:
        five_string = "5"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#int"
        )
        self.assertEqual(five, 5)

    def test_long(self) -> None:
        five_string = "5"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#long"
        )
        self.assertEqual(five, 5)

    def test_integer(self) -> None:
        five_string = "5"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#integer"
        )
        self.assertEqual(five, 5)

    def test_float(self) -> None:
        five_string = "5.0"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#float"
        )
        self.assertEqual(five, 5.0)

    def test_double(self) -> None:
        five_string = "5.0"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#double"
        )
        self.assertEqual(five, 5.0)

    def test_decimal(self) -> None:
        five_string = "5.0"
        five = semantic_type_to_python_object(
            five_string, "http://www.w3.org/2001/XMLSchema#decimal"
        )
        self.assertEqual(five, 5.0)




# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import logging
import time
import signal
from enum import Enum
from uuid import uuid4

from .quad_store import Quad, QuadStore

logger = logging.getLogger(__name__)


class Timeout:
    """
    Helper class for timeouts

    Example usage:
        try:
            with Timeout(seconds=5, error_message="tsup"):
                print("printed")
                time.sleep(10)
                print("never printed")
        except TimeoutError as e:
            print(str(e))
        print("after")
    ```
    """

    def __init__(self, seconds=1, error_message="Timeout") -> None:
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame) -> None:
        raise TimeoutError(self.error_message)

    def __enter__(self) -> None:
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class OrchestrationState(Enum):
    SUCCESS = "SUCCESS"             # success means succeess!
    FAIL = "FAIL"                   # fail indicates the IM sent back a failed state
    INACTIVE = "INACTIVE"           # inactive means that the IM isn't running or hasn't run recently
    IN_PROGRESS = "IN_PROGRESS"     # in_progress means that IM is currently running
    ERROR = "ERROR"                 # error indicates there was some internal error, preventing either a success or failure


class IngestionComponentStatus:
    name_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#componentName"  # noqa
    status_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#componentStatus"  # noqa
    succeeded_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#componentSucceeded"  # noqa
    index_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#index"  # noqa
    start_time_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#ingestStartTime"  # noqa
    end_time_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#ingestEndTime"  # noqa
    percent_complete_pred = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#percentComplete"  # noqa

    true_string = "true"

    def __init__(self, uri: str, name: str,
                 status: str, did_succeed: bool,
                 start_time: str, end_time: str,
                 index: str, percent_complete: str) -> None:

        self.uri = uri
        self.name = name
        self.status = status
        self.did_succeed = did_succeed
        self.start_time = start_time
        self.end_time = end_time
        self.index = index
        self.percent_complete = percent_complete

    @staticmethod
    def from_quadstore(quadstore, uri):

        def get_featured_quad(quadstore, pred):
            """Given a quadstore and a predicate, returns
            the object of the statement with the predicate
            if there is one, returns the empty
            string if there's no predicate and raises
            an error if there's more than one statement
            with the predicate
            """

            quads = quadstore.filter(predicate=pred).as_list()
            if len(quads) == 0:
                result = ""
            elif len(quads) == 1:
                result = quads[0].obj
            else:
                raise RuntimeError("Unable to determine the name of the "
                                   "ingestion component")
            return result

        name = get_featured_quad(quadstore, IngestionComponentStatus.name_pred)
        status = get_featured_quad(quadstore, IngestionComponentStatus.status_pred)
        start_time = get_featured_quad(quadstore, IngestionComponentStatus.start_time_pred)
        end_time = get_featured_quad(quadstore, IngestionComponentStatus.end_time_pred)
        index = get_featured_quad(quadstore, IngestionComponentStatus.index_pred)
        percent_complete = get_featured_quad(quadstore, IngestionComponentStatus.percent_complete_pred)

        success_quad_pred = IngestionComponentStatus.succeeded_pred
        succeeded_quads = quadstore.filter(predicate=success_quad_pred).as_list()
        if len(succeeded_quads) != 1:
            did_succeed = False
        else:
            did_succeed = succeeded_quads[0].obj == IngestionComponentStatus.true_string

        return IngestionComponentStatus(uri, name, status, did_succeed, start_time, end_time, index, percent_complete)

    def __eq__(self, other):
        return (
            self.uri, self.name, self.status, self.did_succeed,
            self.start_time, self.end_time, self.index,
            self.percent_complete
        ) == (
            other.uri, other.name, other.status, other.did_succeed,
            other.start_time, other.end_time, other.index,
            other.percent_complete
        )

    def __str__(self):
        return f"<IngestionComponentStatus name='{self.name}' status='{self.status}' did_succeed='{self.did_succeed}' start_time='{self.start_time}' end_time='{self.end_time}' index='{self.index}' percent_complete='{self.percent_complete}'>"

    def __repr__(self):
        return str(self)


class RunStatus:
    """
    A class that provide information about the RunStatus.

    The overall status of the run is stored in self.status.

    Each ingestion component, each of which corresponds to a load service,
    has its status parsed and stored an IngestionComponentStatus object that's
    in the list self.ingestion_components
    """

    TYPE_PRED = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
    ING_COMPONENT_TYPE = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#IngestionComponent"
    ING_COMPONENT_PRED = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#ingestSucceeded"
    true_string = "true"

    completed_statuses = [
        OrchestrationState.SUCCESS, OrchestrationState.FAIL
    ]

    def __init__(self, status, ingestion_components):
        self.status = status
        self.ingestion_components = ingestion_components

    @staticmethod
    def from_quadstore(quadstore):
        """
        Construct a run status object by parsing a quadstore with results from
        checking the status of an ingestion
        Args:
            quadstore: A quadstore that's the result of a status check of
                an ingestion

        Returns:
            A RunStatus object that's a parsed, easier-to-use form of the
            quadstore
        """

        if len(quadstore) == 0:
            raise ValueError("Unable to parse empty quadstore into RunStatus")

        # First, make a set of the all of the subjects in the quadstore
        # that are of type ingestion componenent. Each load service will
        # correspond to one ingestion component
        ingestion_component_subs = set(
            q.sub for q in
            quadstore.filter(predicate=RunStatus.TYPE_PRED,
                             statement_object=RunStatus.ING_COMPONENT_TYPE)
        )

        # Second, loop over each ingestion component subject and create an
        # ingestion component object
        ing_comps = []
        for ingestion_component_sub in ingestion_component_subs:
            ic_quadstore = quadstore.filter(subject=ingestion_component_sub)
            ing_comp = IngestionComponentStatus.from_quadstore(ic_quadstore, ingestion_component_sub)
            ing_comps.append(ing_comp)

        # Third, determine the overall status of the ingestion
        is_complete_qs = quadstore.filter(predicate=RunStatus.ING_COMPONENT_PRED)
        if len(is_complete_qs) == 0:
            status = OrchestrationState.IN_PROGRESS
        elif len(is_complete_qs) > 1:
            status = OrchestrationState.ERROR
        else:
            if is_complete_qs.as_list()[0].obj == RunStatus.true_string:
                status = OrchestrationState.SUCCESS
            else:
                status = OrchestrationState.FAIL

        return RunStatus(status, ing_comps)

    @property
    def is_complete(self):
        return self.status in self.completed_statuses

    @property
    def was_successful(self):
        return self.status == OrchestrationState.SUCCESS


class OrchestrationWrapper:
    """A class for interacting with an orchestration service and one of its
    ingest managers

    Usage:
        anzo_client = AnzoClient(
            "anzo.cambridgesemantics.com", "8080", "username", "password"
        )

        osm = OrchestrationWrapper(anzo_client, im_uri, load_services)

        osm.start()
        run_status = osm.block_until_complete(2000)
    """

    TYPE_PRED = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
    LOAD_SERVICE_PRED = "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#ingestMgrLoadService"

    STATUS_CHECK_OP = "#getIngestManagerStatuses"
    TRIGGER_SERVICE_OP = "#triggerIngestMgr"

    DEFAULT_ORCH_SERVICE_URI = "http://cambridgesemantics.com/service/OrchestrationService"

    def __init__(self, anzo_client, ingest_manager, load_services, orchestration_uri=""):
        """
        Constructs an OrchestrationWrapper object

        Args:
            anzo_client: AnzoClient object for interacting with anzo
            ingest_manager: URI of the ingest manager
            load_services: List of the load service URIs to execute. If the
                list is empty, then the default load services will be executed
            orchestration_uri (optional): URI of the orchestration service.
                In most cases, this does not need to be provided and the default
                will be used.

        Returns:
            An OrchestrationWrapper object
        """

        self.anzo_client = anzo_client
        self.ingest_manager = ingest_manager
        self.load_services = load_services
        self.named_graph = f"urn://ingestManager/{uuid4()}"

        if orchestration_uri:
            self.orchestration_uri = orchestration_uri
        else:
            self.orchestration_uri = self.DEFAULT_ORCH_SERVICE_URI

        self.status_check_uri = self.orchestration_uri + self.STATUS_CHECK_OP
        self.trigger_service_uri = self.orchestration_uri + self.TRIGGER_SERVICE_OP
        self.poll_interval_seconds = 60  # in seconds


    def start(self):
        service_request = self._build_request(
            self.ingest_manager, self.load_services, self.named_graph
        )

        trigger_response = self.anzo_client.execute_semantic_service(
            self.trigger_service_uri, service_request
        )

        if len(trigger_response) == 0:
            raise RuntimeError("Error triggering the ingest manager")

        # TODO: Question -> is there anything to check here?

    def poll_status(self):
        """Polls for the status of the ingestion and returns a RunStatus
        object with the information about the current status

        Returns:
            A RunStatus object
        """

        max_tries = 3
        num_tries = 0
        while num_tries < max_tries:
            try:
                # Build the request and execute the semantic service to poll the
                # ingest manager status.
                service_request = self._build_request(
                    self.ingest_manager, [], self.named_graph
                )

                poll_response = self.anzo_client.execute_semantic_service(
                    self.status_check_uri, service_request
                )

                msg = (
                    "Response from Ingest Manager status check:"
                    f"\n{poll_response.to_trig_string()}"
                )
                logger.debug(msg)

                # Parse the current status and return it
                run_status = RunStatus.from_quadstore(poll_response)
                return run_status
            except Exception as e:
                logger.warning(
                    "Unable to poll for orchestration service "
                    f"status on attempt {num_tries}"
                )
                logger.exception(e)
                num_tries += 1

        msg = (
            "Unable to poll for orchestration service status "
            f"after {num_tries} attempts"
        )
        logger.error(msg)
        raise RuntimeError(msg)

    def block_until_complete(self, timeout_seconds):
        """
        Blocks the python code until the current run completes. After
        completion, a RunStatus object is returned with information about
        the run.

        Raises a TimeoutError if there's a timeout.

        Args:
            timeout_seconds: The number of seconds to block until timing out

        Returns:
            A RunStatus object with information about the run
        """

        with Timeout(seconds=timeout_seconds):
            run_status = self.poll_status()
            while not run_status.is_complete:
                logger.debug(f"Current status is {run_status.status}. Sleeping.")
                time.sleep(self.poll_interval_seconds)
                run_status = self.poll_status()
            return run_status

    def _build_request(self, ingest_manager, load_services, graph):
        """
        Constructs a request graph for executing semantic services with an
        ingest manager and an optional number of load services.

        No load services can be provided by passing an empty list.

        Example:
            request = self._build_request(
                "urn://im", ["urn://ls1", urn://ls2"], "urn://graph"
            )

        Args:
            ingest_manager: a single uri for an ingest manager
            load_services: a list of load service URIs
            graph: unique named graph for the request

        Returns:
            Quadstore with the request
        """
        # This method will be build up a set of Quads to construct a quadstore
        quads = set()

        # First, add type statements for the Ingest Manager and the Request Graph
        graph_type_objs = [
            "http://openanzo.org/ontologies/2008/07/ASDL/Orchestration#ASDLNode",
            "http://openanzo.org/ontologies/2008/07/RemoteConnection#RemoteAnzoConnectionConfiguration",
            "http://openanzo.org/ontologies/2008/07/System#Credentials",
            "http://openanzo.org/ontologies/2008/07/System#NetworkConnection"
        ]

        im_type_objs = [
            "http://openanzo.org/ontologies/2008/07/ASDL/IngestManager#IngestManagerConfiguration",
            "http://openanzo.org/ontologies/2008/07/ASDL/Orchestration#ASDLNodeController"
        ]

        for type_obj in graph_type_objs:
            quads.add(Quad(graph, self.TYPE_PRED, type_obj, graph))

        for type_obj in im_type_objs:
            quads.add(Quad(ingest_manager, self.TYPE_PRED, type_obj, graph))

        # Second, add a quad pointing to the ingest manager
        quads.add(Quad(
            graph, "http://openanzo.org/ontologies/2008/07/ASDL/Orchestration#controllerServiceConfiguration",
            ingest_manager, graph
        ))

        # Third, if specific load services are specified, then include a quad pointing to each one
        for load_service in load_services:
            quads.add(Quad(
                ingest_manager, self.LOAD_SERVICE_PRED,
                load_service, graph
            ))

        return QuadStore(quads)

    def __str__(self):
        return (
            "<OrchestrationWrapper "
            f"orchestration_uri='{self.orchestration_uri}', "
            f"ingest_manager_uri='{self.ingest_manager}'>"
        )

    def __repr__(self):
        return str(self)



#!/usr/bin/env python

"""
/*******************************************************************************
 * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
 * All rights reserved.
 * 
 * Contributors:
 *     Cambridge Semantics Incorporated
 *******************************************************************************
 """

from setuptools import setup, find_packages

#with open('readme.md') as readme_file:
 #   readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = ['urllib3>=1.25.11', 'requests>=2.23.0', 'rdflib>=5.0.0']

setup_requirements = []

test_requirements = []

setup(
    author="Tommy Fang, William Ades, Curtis Galione, Andrew Parisi, Alex Ledger",
    author_email='info@cambridgesemantics.com',
    python_requires='>=3.5',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    description="PyAnzo is a library for interacting with Anzo",

    entry_points={
        'console_scripts': [],
    },
    install_requires=requirements,
#    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",

    keywords='pyanzo',
    name='pyanzo',
    packages=find_packages(include=['pyanzo', 'pyanzo.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='',
    version='3.3.4',
    zip_safe=False,
)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
import os
from pyanzo import AnzoClient, QuadStore, Quad

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestAnzoClientQueryJournalBase(unittest.TestCase):
    journal_query1 = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>

        SELECT ?title
        WHERE {
            GRAPH ?g {
                ?s a <http://cambridgesemantics.com/ontologies/Graphmarts#Graphmart> ;
                    dc:title ?title .
            }
            FILTER(?g = <http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
        }
    """  # noqa

    journal_query2 = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        SELECT ?title
        WHERE {
            GRAPH <http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7> {
                ?s a <http://cambridgesemantics.com/ontologies/Graphmarts#Graphmart> ;
                    dc:title ?title .
            }
        }
    """  # noqa

    journal_query3 = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        SELECT ?title
        WHERE {
            GRAPH ?g {
                ?s a <http://cambridgesemantics.com/ontologies/Graphmarts#Graphmart> ;
                    dc:title ?title .
            }
            VALUES (?g) {
                (<http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
            }
        }
    """  # noqa

    journal_query4 = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        SELECT ?title
        WHERE {
            ?s a <http://cambridgesemantics.com/ontologies/Graphmarts#Graphmart> ;
                dc:title ?title .
            FILTER(?s = <http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
        }
    """  # noqa

    journal_queries = [
        journal_query1, journal_query2, journal_query3, journal_query4
    ]

    journal_query_results = [["PyAnzo Graphmart"]]

    journal_query_target_graph_list = [
        "http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7"  # noqa
    ]

    simple_journal_query_file = os.path.join(
        "tests", "test_assets", "simple_journal_query.rq"
    )

    malformed_query_file = os.path.join(
        "tests", "test_assets", "malformed_journal_query.rq"
    )

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def test_journal_query_with_query_string_and_file(self) -> None:
        self.assertRaises(
            ValueError, self.anzo_client.query_journal,
            query_string=self.journal_query1,
            query_file=self.simple_journal_query_file
        )

    def test_journal_query(self) -> None:
        for journal_query in self.journal_queries:
            result = self.anzo_client.query_journal(journal_query)
            self.assertEqual(
                result.as_table_results().as_list(),
                self.journal_query_results
            )

    def test_journal_query_with_graph_specified(self) -> None:
        for journal_query in self.journal_queries:
            result = self.anzo_client.query_journal(
                journal_query,
                named_graphs=self.journal_query_target_graph_list
            )
            self.assertEqual(
                result.as_table_results().as_list(),
                self.journal_query_results
            )

    def test_bad_journal_query(self) -> None:
        query = "SELECT ?title"
        self.assertRaises(RuntimeError, self.anzo_client.query_journal, query)

    def test_simple_journal_query_with_file(self) -> None:
        result = self.anzo_client.query_journal(
            query_file=self.simple_journal_query_file
        )
        self.assertEqual(
            result.as_table_results().as_list(), self.journal_query_results
        )

    def test_journal_query_with_bad_file(self) -> None:
        self.assertRaises(
            RuntimeError, self.anzo_client.query_journal,
            query_file="abc123.123abc"
        )

    def test_malformed_journal_query_with_bad_file(self) -> None:
        self.assertRaises(
            RuntimeError, self.anzo_client.query_journal,
            query_file=self.malformed_query_file
        )


class TestAnzoClientQueryJournalTwoGraphs(unittest.TestCase):
    journal_query_two_graphs = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        SELECT ?title
        WHERE {
            GRAPH ?g {
                ?s dc:title ?title .
            }

            VALUES (?g) {
                (<http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
                (<http://csi.com/FileBasedLinkedDataSet/bd05ae950df39757dfbb5a6fbc5c925e>)
            }
        }
        ORDER BY ?title
    """  # noqa

    journal_query_two_graphs_results = [["PyAnzo Dataset"], ["PyAnzo Graphmart"]]  # noqa

    journal_query_two_graphs_list = [
        "http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7",  # noqa
        "http://csi.com/FileBasedLinkedDataSet/bd05ae950df39757dfbb5a6fbc5c925e"  # noqa
    ]

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def test_with_default_graphs(self) -> None:
        result = self.anzo_client.query_journal(
            query_string=self.journal_query_two_graphs
        )
        self.assertEqual(
            result.as_table_results().as_list(),
            self.journal_query_two_graphs_results
        )

    def test_with_two_graphs_specified(self) -> None:
        result = self.anzo_client.query_journal(
            query_string=self.journal_query_two_graphs,
            named_graphs=self.journal_query_two_graphs_list
        )
        self.assertEqual(
            result.as_table_results().as_list(),
            self.journal_query_two_graphs_results
        )

    def test_with_one_graph_specified(self) -> None:
        result = self.anzo_client.query_journal(
            query_string=self.journal_query_two_graphs,
            named_graphs=[self.journal_query_two_graphs_list[0]]
        )
        self.assertNotEqual(
            result.as_table_results().as_list(),
            self.journal_query_two_graphs_results
        )


class TestAnzoClientQueryJournalConstruct(unittest.TestCase):
    query = """
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        CONSTRUCT {
            GRAPH ?g {
                ?s <urn://predicate> ?title .
            }
        }
        WHERE {
            GRAPH ?g {
                ?s a <http://cambridgesemantics.com/ontologies/Graphmarts#Graphmart> ;
                    dc:title ?title .
            }
            FILTER(?g = <http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7>)
        }
    """  # noqa

    set_of_quads = {
        Quad(
            sub='http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7',  # noqa
            pred='urn://predicate',
            obj='PyAnzo Graphmart',
            graph='http://cambridgesemantics.com/Graphmart/9da211618a15476daa10cead2292d8e7',  # noqa
            sub_type="uri",
            obj_type="literal",
        )
    }

    expected_quad_store = QuadStore(set_of_quads)

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            server=SERVER, username=USERNAME,
            password=PASSWORD, port=PORT
        )

    def test_construct_query(self) -> None:
        result = self.anzo_client.query_journal(query_string=self.query)
        self.assertEqual(result.as_quad_store(), self.expected_quad_store)



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

from pyanzo.quad_store import QuadStore, Quad
import unittest
import datetime
import urllib.parse

from pyanzo import AnzoClient
from pyanzo.lds_manager import LDSManager, DNG_PRED
from pyanzo import uris
from pyanzo.uris import SYSTEM_DATASOURCE_URI

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD
)


class TestLDSManager(unittest.TestCase):
    """
    Test class for LDS manager. On class initialization, an LDS is created
    with a timestampped title and the "quads_to_add" are added to the LDS.
    The class's lds_mgr manages this LDS. All tests should assume that they
    get the managed LDS in this same state, since tearDown() reverts the LDS
    back to this starting state.
    """
    
    # Paramterized values for the tests. This value is overwritten by the
    # values in the decorator

    ng_1_uri = "http://csi.test.com/"
    ng_2_uri = "http://csi.test.com/graph2"
    ng_3_uri = "http://csi.test.com/graph3"

    # Quads to add to LDS
    quads_to_add = QuadStore({
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate3",
             "7",
             ng_2_uri,
             obj_type="literal",
             obj_data_type="http://www.w3.org/2001/XMLSchema#integer"),
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate2",
             "http://csi.test.com/object2",
             ng_1_uri)
    })

    # different quads to add after clearing LDS
    new_quads_to_add = QuadStore({
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate3",
             "8",
             ng_3_uri,
             obj_type="literal",
             obj_data_type="http://www.w3.org/2001/XMLSchema#integer"),
        Quad("http://csi.test.com/subject2",
             "http://csi.test.com/predicate2",
             "http://csi.test.com/object2",
             ng_1_uri)
    })

    @classmethod
    def setUpClass(cls) -> None:
        cls.az_client = AnzoClient(
            server=SERVER, username=USERNAME, password=PASSWORD, port=PORT
        )

        now = datetime.datetime.now().strftime("%m/%d/%y - %H:%M:%S")
        cls.lds_title = "test LDS %s" % now
        cls.lds_mgr = LDSManager.create_lds(anzo_client=cls.az_client,
                                            title=cls.lds_title,
                                            statements=cls.quads_to_add,
                                            ontologies=["urn://ont"])
        cls.lds_uri = cls.lds_mgr.lds_uri

    @classmethod
    def tearDownClass(cls):
        # remove dataset created for test
        cls.lds_mgr.delete_lds()

    def tearDown(self) -> None:
        # reset back to "normal" state for further tests
        self.lds_mgr.overwrite_lds(self.quads_to_add)

    def test_created(self):
        # dataset was created in set up
        # test that the right artifacts are there

        qs = self.assert_lds_exists()

        # lds should have two NGs
        self.assertEqual(len(qs.filter(predicate=DNG_PRED)), 2)
        onto_pred = uris.NAMESPACE["cs-ld"] + "ontology"
        # lds should have one onto
        self.assertEqual(len(qs.filter(predicate=onto_pred)), 1)
        # lds should have correct title
        title = qs.filter(predicate=uris.NAMESPACE["dc"] + "title",
                          statement_object=self.lds_title)
        self.quadstore_not_empty(title)

        # actual graphs should exist and not be empty
        self.quadstore_not_empty(self.az_client.get(self.ng_1_uri))
        self.quadstore_not_empty(self.az_client.get(self.ng_2_uri))

    def test_clear_add(self):
        self.lds_mgr.clear_lds()
        qs = self.assert_lds_exists()

        # lds should have 0 NGs
        self.assertEqual(len(qs.filter(predicate=DNG_PRED)), 0)

        # actual graphs should not exist/be empty
        self.graph_empty(self.ng_1_uri)
        self.graph_empty(self.ng_2_uri)

        # add data back
        self.lds_mgr.add_to_lds(self.quads_to_add)

        # actual graphs should now exist and not be empty
        self.quadstore_not_empty(self.az_client.get(self.ng_1_uri))
        self.quadstore_not_empty(self.az_client.get(self.ng_2_uri))

    def test_overwrite(self):
        self.lds_mgr.overwrite_lds(self.new_quads_to_add)

        qs = self.assert_lds_exists()

        # lds should have two NGs
        self.assertEqual(len(qs.filter(predicate=DNG_PRED)), 2)

        # ng 1 and 3 should exist and be not empty
        self.quadstore_not_empty(self.az_client.get(self.ng_1_uri))
        self.quadstore_not_empty(self.az_client.get(self.ng_3_uri))

        # ng2 should not exist/be empty
        self.graph_empty(self.ng_2_uri)

    def assert_lds_exists(self) -> QuadStore:
        # first look at cat entry graph
        qs = self.az_client.get(cat_entry_graph(self.lds_uri))
        self.quadstore_not_empty(qs)
        online = qs.filter(predicate=uris.NAMESPACE["cs-ld"] + "online",
                           statement_object="true")
        self.quadstore_not_empty(online)
        title = qs.filter(predicate=uris.NAMESPACE["dc"] + "title",
                          statement_object=self.lds_title)
        self.quadstore_not_empty(title)

        # then look at lds graph itself
        qs = self.az_client.get(self.lds_uri)
        self.quadstore_not_empty(qs)
        return qs

    def quadstore_not_empty(self, quad_store: QuadStore):
        self.assertGreater(len(quad_store), 0)

    def graph_empty(self, graph_uri: str):
        self.assertRaises(RuntimeError, self.az_client.get, graph_uri)


def cat_entry_graph(lds_uri: str,
                    datasource_uri: str = SYSTEM_DATASOURCE_URI) -> str:
    """Utility method to construct the catalog entry URI
    for the given LDS and datasource URIs.

    Args:
        lds_uri: URI of the LDS
        datasource_uri: (optional) URI of the datasource.
            Defaults to the system datasource.

    Returns:
        The catalog entry URI.
    """
    encoded = urllib.parse.quote("[%s]@[%s]" %
                                 (lds_uri, datasource_uri), safe="")
    return "http://openanzo.org/catEntry(%s)" % encoded



# /*******************************************************************************
#  * Copyright (c) 2019 - 2022 Cambridge Semantics Incorporated.
#  * All rights reserved.
#  * 
#  * Contributors:
#  *     Cambridge Semantics Incorporated
#  *******************************************************************************/

import unittest
from pyanzo import AnzoClient, Quad, QuadStore

from .test_common import (
    SERVER,
    PORT,
    USERNAME,
    PASSWORD,
)


class TestAnzoClientGet(unittest.TestCase):
    graph_uri = "http://cambridgesemantics.com/namedgraph/1"
    md_graph_uri = "http://openanzo.org/metadataGraphs(http%3A%2F%2Fcambridgesemantics.com%2Fnamedgraph%2F1)"  # noqa

    expected_quad = Quad(
        sub='http://cambridgesemantics.com/namedgraph/1/sub',
        pred='http://cambridgesemantics.com/namedgraph/1/pred',
        obj='http://cambridgesemantics.com/namedgraph/1/obj',
        graph='http://cambridgesemantics.com/namedgraph/1'
    )

    expected_quadstore = QuadStore({expected_quad})

    def setUp(self) -> None:
        self.anzo_client = AnzoClient(
            SERVER, PORT, username=USERNAME, password=PASSWORD)

    def test_get(self) -> None:
        result = self.anzo_client.get(self.graph_uri)
        result = result.filter(graph=self.expected_quad.graph)
        self.assertEqual(result, self.expected_quadstore)

    def test_get_with_nonexistent_graph(self) -> None:
        graph_uri = "urn://this/graph/doesnt/exist"
        self.assertRaises(RuntimeError, self.anzo_client.get, graph_uri)

    def test_get_with_metadatagraph(self) -> None:
        result = self.anzo_client.get(
            self.graph_uri, include_metadata_graph=True
        )

        md_graph = result.filter(graph=self.md_graph_uri)
        self.assertTrue(len(md_graph) > 3)

        self.assertNotEqual(result, self.expected_quadstore)

        self.assertEqual(
            result.filter(graph=self.expected_quad.graph),
            self.expected_quadstore
        )



"""Library for interacting with Anzo

PyAnzo is a python library for interacting with an AnzoServer

Usage::
    anzo_client = AnzoClient(server="localhost", port="443",
                             username="user", password="password")

    query = "SELECT * WHERE { ?s ?p ?o } LIMIT 10",
    graphmart = "urn://graphmart_uri"
    query_result = anzo_client.query_graphmart(graphmart, query_string=query)
    result_as_list = query_result.as_table_result.as_list()
"""

__author__ = """Cambridge Semantics"""
__email__ = ''
__version__ = '4.0.0'

from .anzo_client import AnzoClient
from .query_result import QueryResult, TableResult
from .anzo_request import AnzoRequestBuilder, AnzoRequest
from .quad_store import Quad, QuadStore, semantic_type_to_python_object
from .uris import ALL_NAMED_GRAPHS_URI, TYPE_PRED

__all__ = [
    'AnzoClient',
    'QueryResult',
    'TableResult',
    'AnzoRequest',
    'AnzoRequestBuilder',
    'LDSManager',
    'Quad',
    'QuadStore',
    'semantic_type_to_python_object',
    'ALL_NAMED_GRAPHS_URI',
    'TYPE_PRED',
]
